import numpy as np
import pygame
import random
import math
import sys
from PyQt5.QtWidgets import QApplication, QMainWindow, QLabel, QVBoxLayout, QWidget, QDialog, QComboBox, \
    QDialogButtonBox
from matplotlib import pyplot as plt
from tabulate import tabulate

# Velikost okna
window_width = 800
window_height = 600
# Velikost statistik
stats_width = 400
stats_height = 600
# Nastavení barev
bg_color = (255, 255, 255)
target_color = (0, 0, 153)
player_color = (0, 0, 0)
# Nastavení obrázků
background_image = pygame.image.load("pictures/background.png")
obstacle_image = pygame.image.load("pictures/obstacle.png")
target_image = pygame.image.load("pictures/target.png")


# Okno pro nastavení simulace
class ConfigurationDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Nastavení")
        self.setModal(True)

        self.crossover_combo_box = QComboBox(self)
        self.crossover_combo_box.addItem("Jednobodové křížení")
        self.crossover_combo_box.addItem("Dvoubodové křížení")
        self.crossover_combo_box.addItem("Uniformní křížení")

        self.population_combo_box = QComboBox(self)
        self.population_combo_box.addItem("600")
        self.population_combo_box.addItem("800")
        self.population_combo_box.addItem("1000")

        self.selection_combo_box = QComboBox(self)
        self.selection_combo_box.addItem("Ruletový výběr")
        self.selection_combo_box.addItem("Turnajový výběr")

        self.obstacle_combo_box = QComboBox(self)
        self.obstacle_combo_box.addItem("Jednoduchá překážka")
        self.obstacle_combo_box.addItem("Složité překážky")

        button_box = QDialogButtonBox(QDialogButtonBox.Ok)
        button_box.accepted.connect(self.accept)

        layout = QVBoxLayout(self)
        layout.addWidget(QLabel("Vyberte druh křížení"))
        layout.addWidget(self.crossover_combo_box)
        layout.addWidget(QLabel("Zvolte velikost populace"))
        layout.addWidget(self.population_combo_box)
        layout.addWidget(QLabel("Vyberte druh selekce"))
        layout.addWidget(self.selection_combo_box)
        layout.addWidget(QLabel("Vyberte druh překážek"))
        layout.addWidget(self.obstacle_combo_box)
        layout.addWidget(button_box)

    def get_crossover_type(self):
        return self.crossover_combo_box.currentText()

    def get_population_size(self):
        return int(self.population_combo_box.currentText())

    def get_selection_type(self):
        return self.selection_combo_box.currentText()

    def get_obstacle_type(self):
        return self.obstacle_combo_box.currentText()


# Inicializace okna pro statistiky
class StatsWindow(QMainWindow):
    def __init__(self, population):
        super().__init__()
        self.population = population
        self.crossover_chosen = False
        self.setWindowTitle("Statistiky")
        self.setGeometry(window_width + 885, 420, stats_width, stats_height)  # Umístění okna vedle Pygame okna
        self.layout = QVBoxLayout()
        self.widget = QWidget()
        self.widget.setLayout(self.layout)
        self.setCentralWidget(self.widget)

        self.generation_label = QLabel(self)
        self.generation_label.move(10, 10)
        self.layout.addWidget(self.generation_label)

        self.player_count_label = QLabel(self)
        self.player_count_label.move(10, 40)
        self.layout.addWidget(self.player_count_label)

        self.remaining_time_label = QLabel(self)
        self.remaining_time_label.move(10, 70)
        self.layout.addWidget(self.remaining_time_label)

        self.average_fitness_label = QLabel(self)
        self.average_fitness_label.move(10, 100)
        self.layout.addWidget(self.average_fitness_label)

        self.best_fitness_label = QLabel(self)
        self.best_fitness_label.move(10, 130)
        self.layout.addWidget(self.best_fitness_label)

        self.mutation_label = QLabel(self)
        self.mutation_label.move(10, 160)
        self.layout.addWidget(self.mutation_label)

        self.selection_label = QLabel(self)
        self.selection_label.move(10, 190)
        self.layout.addWidget(self.selection_label)

        self.crossover_label = QLabel(self)
        self.crossover_label.move(10, 220)
        self.layout.addWidget(self.crossover_label)

        self.obstacle_label = QLabel(self)
        self.obstacle_label.move(10, 250)
        self.layout.addWidget(self.obstacle_label)

        self.hint1_label = QLabel(self)
        self.hint1_label.move(10, 280)
        self.layout.addWidget(self.hint1_label)

        self.hint2_label = QLabel(self)
        self.hint2_label.move(10, 310)
        self.layout.addWidget(self.hint2_label)
        self.show()

    # Aktualizace a popis statistik
    def update(self):
        stats_window.generation_label.setText(f"Generace: {population.generation}.")
        stats_window.player_count_label.setText(
            f"Zbývající hráčí: {population.get_alive_player_count()}/{population_size}")
        stats_window.remaining_time_label.setText(f"Zbývající čas: {max(generation_time - current_time, 0):.1f}s")
        stats_window.mutation_label.setText(f"Aktuální míra mutace: {population.mutation_rate * 100} %")
        stats_window.selection_label.setText(f"Aktuální selekce: {population.selection_type}")
        stats_window.crossover_label.setText(f"Aktuální křížení: {population.crossover_type}")
        stats_window.obstacle_label.setText(f"Aktuální překážky: {population.obstacle_type}")
        stats_window.average_fitness_label.setText(f"Průměrná fitness předchozí generace: {population.average_fitness}")
        stats_window.best_fitness_label.setText(f"Nejvyšší fitness v předchozí generaci: {population.best_fitness}")
        stats_window.hint1_label.setText(f"Mezerník pro zrychlení")
        stats_window.hint2_label.setText(f"Escape pro ukončení")


# Cíl
class Target:
    def __init__(self, position, size):
        self.position = position
        self.size = size
        self.initial_position = position

    def reset(self):
        self.position = self.initial_position

    def draw(self):
        # Vykreslení cíle na herní ploše
        target_rect = pygame.Rect(self.position[0] - self.size // 2, self.position[1] - self.size // 2, self.size,
                                  self.size)
        # Změna velikosti obrázku cíle na požadovanou velikost
        target_image_scaled = pygame.transform.scale(target_image, (self.size, self.size))
        # Vykreslení zmenšeného obrázku cíle na herní plochu
        window.blit(target_image_scaled, target_rect)


# Hráč
class Player:
    def __init__(self, genes, min_distance=None, max_distance=None):
        self.x = window_width // 2
        self.y = window_height - 50
        self.velocity = 7
        self.genes = genes
        self.step = 0
        self.is_alive = True
        self.is_finished = False
        self.size = 20
        self.fitness = 0
        self.min_distance = min_distance
        self.max_distance = max_distance
        self.player_image = pygame.transform.scale(pygame.image.load("pictures/player.png"), (self.size, self.size))
        self.positions_history = []

    # Pohyb hráče na základě genetického kódu
    def move(self):
        if not self.is_alive or self.is_finished:  # Kontrola, zda je hráč stále živý nebo již nedosáhl cíle
            return
        angle = self.genes[
            self.step]  # Z genetického kódu hráče se na základě aktuálního kroku získá úhel, pod kterým se bude hráč pohybovat
        new_x = self.x + self.velocity * math.cos(angle)
        new_y = self.y + self.velocity * math.sin(angle)
        # Zamezení hráče pro vystoupení z okna
        if 0 <= new_x <= window_width - 25:
            self.x = new_x
        if 0 <= new_y <= window_height - 25:
            self.y = new_y
        self.positions_history.append((self.x, self.y))  # Přidání pozice do seznamu pro vykreslování heatmapy
        self.step += 1
        if self.step >= len(
                self.genes):  # Pokud hráč provedl všechny kroky daného genetického kódu, prochází svůj genetický kód od začátku
            self.step = 0

    # Vykreslení hráče
    def draw(self):
        window.blit(self.player_image, (self.x, self.y))  # Vykreslení obrázku hráče

    # Detekce kolize hráče s překážkami
    def check_collision(self):
        if not self.is_alive or self.is_finished:
            return
        for obstacle in obstacles:
            if (self.x < obstacle.x + obstacle.width and
                    self.x + 10 > obstacle.x and
                    self.y < obstacle.y + obstacle.height and
                    self.y + 10 > obstacle.y):
                self.is_alive = False  # Hráč umře při kolizi
                population.remaining_players -= 1  # Snížení počtu zbývajících hráčů v populaci
                return

    # Výpočet vzdálenosti hráče od cíle
    def calculate_distance_to_target(self):
        target_x, target_y = target_position
        distance = math.sqrt((self.x - target_x) ** 2 + (self.y - target_y) ** 2)
        return distance

    # Ověření zda se hráč nachází v cíli
    def is_colliding_with_target(self, target_position, target_size):
        target_x, target_y = target_position
        distance = math.sqrt((self.x - target_x) ** 2 + (self.y - target_y) ** 2)
        return distance < (self.size + target_size) / 2

    # Ověření, zda hráč dosáhl cíle a volání konce hry
    def check_target_reached(self, target_position, target_size):
        # Kontrola, zda hráč koliduje s cílem
        if self.is_colliding_with_target(target_position, target_size):
            output_file = "heatmap.png"
            # Vytvoření teplotní mapy a uložení do souboru
            create_heatmap_and_save_to_png(
                population.players, window_width, window_height, output_file,
                target_position=target_position, obstacles=obstacles)
            # Ukončení hry, protože hráč dosáhl cíle
            self.game_over()
        else:
            # Hráč cíle nedosáhl, hra pokračuje dále
            return False

    # Výpočet hodnoty fitness pro daného hráče
    def calculate_fitness(self):
        # Výpočet fitness hodnoty hráče na základě jeho polohy a cíle
        if not self.is_alive:  # Kontrola, zda je hráč naživu
            self.fitness = 0
            return
        target_x, target_y = target_position
        # Výpočet vzdálenosti hráče od cíle
        distance = math.sqrt((self.x - target_x) ** 2 + (self.y - target_y) ** 2)
        # Normalizace vzdálenosti pro hodnoty 0 až 100
        if self.min_distance is not None and self.max_distance is not None:
            normalized_distance = (distance - self.min_distance) / (self.max_distance - self.min_distance)
            # Výpočet fitness na základě normalizované vzdálenosti
            self.fitness = 100 - (normalized_distance * 100)
        else:
            # Pokud nejsou definovány min/max vzdálenosti, použije se inverzní hodnota vzdálenosti
            self.fitness = 1 / distance

    # Resetování hráče na počáteční pozici, výchozí hodnoty pro novou generaci
    def reset(self):
        self.x = window_width // 2
        self.y = window_height - 50
        self.is_alive = True
        self.is_finished = False

    # Ukončení hry a zobrazení informací o průběhu simulace
    def game_over(self):
        font = pygame.font.SysFont(None, 36)
        text = font.render(f"Konec hry, počet generací: {population.generation}", True, (0, 0, 0))
        text_rect = text.get_rect(center=(window_width // 2, window_height // 2))
        window.blit(text, text_rect)
        population.update_fitness()
        population.best_fitness = 100
        population.table_population_size = population.get_alive_player_count()
        population.update_table(population.generation)
        population.display_table()
        print(f"Celkový počet kroků: {self.step}")
        print(f"Druh selekce:  {selection_type}")
        print(f"Druh křížení:  {crossover_type}")
        print(f"Druh překážek:  {obstacle_type}")
        pygame.display.update()
        pygame.time.wait(2000)  # Počkání 2 sekund, než se vypne hra
        sys.exit()


# Překážky
class Obstacle:
    def __init__(self, x, y, width, height, obstacle_type):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.initial_x = x
        self.initial_y = y
        self.obstacle_type = obstacle_type

    def reset(self):
        self.x = self.initial_x
        self.y = self.initial_y

    # Metoda pro kreslení překážek
    def draw(self):
        obstacle_rect = obstacle_image.get_rect()
        obstacle_rect.x = self.x
        obstacle_rect.y = self.y
        obstacle_image_scaled = pygame.transform.scale(obstacle_image, (self.width, self.height))
        window.blit(obstacle_image_scaled, obstacle_rect)

    # Hranice na okrajích okna
    @staticmethod
    def create_border_obstacles():
        obstacle_width = 10
        obstacle_height = 10
        border_obstacles = []

        for i in range(0, window_width, obstacle_width):
            border_obstacles.append(Obstacle(i, 0, obstacle_width, obstacle_height, obstacle_type))
            border_obstacles.append(
                Obstacle(i, window_height - obstacle_height, obstacle_width, obstacle_height, obstacle_type))

        for i in range(0, window_height, obstacle_height):
            border_obstacles.append(Obstacle(0, i, obstacle_width, obstacle_height, obstacle_type))
            border_obstacles.append(
                Obstacle(window_width - obstacle_width, i, obstacle_width, obstacle_height, obstacle_type))

        return border_obstacles


# Populace hráčů
class Population:
    def __init__(self, size, elite_percentage, gene_length, stats_window, min_distance, max_distance):
        self.players = [Player([random.uniform(0, 2 * math.pi) for _ in range(gene_length)], min_distance, max_distance)
                        for _ in range(size)]
        self.generation = 1
        self.remaining_players = size
        self.elite_percentage = elite_percentage
        self.gene_length = gene_length
        self.mutation_rate = 0
        self.average_fitness = 0
        self.best_fitness = 0
        self.previous_average_fitness = 0
        self.table_population_size = 0
        self.low_mutation_rate = 0.05
        self.high_mutation_rate = 0.1
        self.min_distance = min_distance
        self.max_distance = max_distance
        self.obstacle_type = obstacle_type
        self.crossover_type = None  # Výchozí hodnota pro druh křížení
        self.selection_type = None  # Výchozí hodnota pro selekci
        self.stats_window = stats_window  # Uložení instance StatsWindow
        self.table = []  # Tabulka pro ukládání hodnot generací, průměrné fitness a nejlepší fitness

    def start_game(self, crossover_type, population_size, selection_type):
        self.crossover_type = crossover_type
        self.selection_type = selection_type
        self.population_size = int(population_size)

    # Slouží k atualizaci stavu hráčů
    def update(self):
        for player in self.players:
            if player.is_alive:
                player.move()
                player.check_collision()
                player.check_target_reached(target_position, target_size)
                if not player.is_alive:
                    self.remaining_players -= 1  # Snížení počtu zbývajících hráčů v populaci

    # Vykreslení hráčů, kteří jsou stále na živu
    def draw(self):
        for player in self.players:
            if player.is_alive:
                player.draw()

    # Metody evolučních algoritmů
    # Metoda pro výběr selekce
    def selection(self):
        if self.selection_type == "Ruletový výběr":
            return self.roulette_selection()
        elif self.selection_type == "Turnajový výběr":
            tournament_size = 8
            return self.tournament_selection(tournament_size)
        else:
            raise ValueError("Neplatný druh selekce")

    # Selekce s využitím ruletového výběru
    def roulette_selection(self):
        total_fitness = sum(player.fitness for player in self.players)  # Celková fitness populace
        pick = random.uniform(0, total_fitness)  # Náhodný bod na ruletě
        current = 0
        # Procházení hráčů a postupné sčítání fitness
        for player in self.players:
            current += player.fitness
            # Výběr hráče, pokud aktuální suma fitness je větší než náhodný bod na ruletě
            if current > pick and player.is_alive:  # Výběr pouze z živých hráčů
                return player

    # Selekce s využitím turnajového výběru
    def tournament_selection(self, tournament_size):
        alive_players = [player for player in self.players if player.is_alive]  # Vytvoření seznamu živých hráčů
        tournament_players = random.sample(alive_players, tournament_size)  # Náhodný výběr hráčů do turnaje
        tournament_players.sort(key=lambda player: player.fitness, reverse=True)  # Seřazení hráčů podle fitness
        return tournament_players[0]  # Vrátíme hráče s nejvyšší fitness

    # Metoda pro výběr křížení
    def crossover(self, parent1, parent2):
        if self.crossover_type == "Jednobodové křížení":
            return self.single_point_crossover(parent1, parent2)
        elif self.crossover_type == "Dvoubodové křížení":
            return self.two_point_crossover(parent1, parent2)
        elif self.crossover_type == "Uniformní křížení":
            return self.uniform_crossover(parent1, parent2)
        else:
            raise ValueError("Neplatný druh křížení")

    # Uniformní křížení
    @staticmethod
    def uniform_crossover(parent1, parent2):
        child_genes = []
        # Procházení genů obou rodičů postupně
        for gene1, gene2 in zip(parent1.genes, parent2.genes):
            # Náhodný výběr jednoho z genů z rodičů pro potomka
            child_gene = random.choice([gene1, gene2])
            child_genes.append(child_gene)
        # Vytvoření nového potomka s kombinovanými geny
        child = Player(child_genes)
        return child

    # Jednobodové křížení
    @staticmethod
    def single_point_crossover(parent1, parent2):
        # Náhodně zvolený bod pro křížení genů
        crossover_point = random.randint(0, len(parent1.genes))
        # Kombinace genů z rodičů v bodě křížení
        genes = parent1.genes[:crossover_point] + parent2.genes[crossover_point:]
        # Vytvoření nového potomka s kombinovanými geny
        new_child = Player(genes=genes)
        return new_child

    # Dvoubodové křížení
    @staticmethod
    def two_point_crossover(parent1, parent2):
        # Zajištění, že první rodič má menší délku genetického kódu, pokud tomu tak není, rodiče jsou prohozeni
        if len(parent1.genes) > len(parent2.genes):
            parent1, parent2 = parent2, parent1
        size = min(len(parent1.genes), len(parent2.genes))  # Určení délky kódu pro křížení
        # Dva náhodné body křížení
        crossover_point1 = random.randint(0, size)
        crossover_point2 = random.randint(0, size)
        # Body křížení seřazené vzestupně
        crossover_start = min(crossover_point1, crossover_point2)
        crossover_end = max(crossover_point1, crossover_point2)
        child = Player(genes=[])
        # Geny před prvním bodem křížení
        child.genes[:crossover_start] = parent1.genes[:crossover_start]
        # Geny mezi 2 body křížení
        child.genes[crossover_start:crossover_end] = parent2.genes[crossover_start:crossover_end]
        # Geny po 2. bodu křížení
        child.genes[crossover_end:] = parent1.genes[crossover_end:]
        return child

    # Mutace
    def mutation(self, player):
        # Kontrola, zda je průměrná fitness větší než předchozí fitness
        if self.average_fitness > self.previous_average_fitness:
            self.mutation_rate = self.low_mutation_rate  # Snížení míry mutace
        else:
            self.mutation_rate = self.high_mutation_rate  # Zvýšení míry mutace

        # Procházení všech genů hráče
        for i in range(len(player.genes)):
            # Pravděpodobnost mutace genu
            if random.random() < self.mutation_rate:
                # Přidání náhodné hodnoty v rozsahu <-1.5, 1.5> k aktuálnímu genu
                player.genes[i] += random.uniform(-1.5, 1.5)
                # Zajištění, aby hodnota genu byla v rozmezí <0, 2*pi>
                player.genes[i] %= 2 * math.pi

    # Vytvoření další generace hráčů v populaci
    def next_generation(self):
        self.update_fitness()  # Aktualizace fitness hodnot pro všechny hráče v současné generaci
        if any(player.fitness is None for player in
               self.players):  # Prověření zda některý z hráčů nemá vypočátanou fitness hodnotu
            return
        self.players.sort(key=lambda player: player.fitness, reverse=True)  # Seřazení hráčů podle fitness
        new_players = []  # Seznam nových hráčů
        elite_count = int(
            self.elite_percentage * len(self.players))  # Elitismus zachování top 10% hráčů z předchozí generace
        new_players.extend(self.players[:elite_count])  # Přidání elitních hráčů
        remaining_count = len(self.players) - elite_count  # Zbývající počet hráčů
        for _ in range(remaining_count):  # Vytváření nové generace
            parent1 = self.selection()  # Selekce první rodič
            parent2 = self.selection()  # Selekce druhý rodič
            child = self.crossover(parent1, parent2)  # Křížení, vytvoření potomka
            self.mutation(child)  # Mutace na potomkovi
            new_players.append(child)  # Přidání potomka do nové generace
        self.players = new_players
        self.remaining_players = len(self.players)
        # Přidání hodnot do tabulky pro aktuální generaci
        population.update_table(self.generation)  # Aktualizace tabulky o populaci pro aktuální generaci
        population.display_table()  # Zobrazení tabulky
        self.generation += 1
        for player in self.players:
            player.reset()
            player.is_alive = True
            player.fitness = None
            player.min_distance = self.min_distance
            player.max_distance = self.max_distance

    # Výpočet fitness pro všechny hráče v populaci -
    def update_fitness(self):
        for player in self.players:
            player.calculate_fitness()
        self.previous_average_fitness = self.average_fitness
        self.average_fitness = sum(player.fitness for player in self.players) / len(self.players)
        self.best_fitness = max(player.fitness for player in self.players)

    # Zkontroluje, zda jsou všichni hráči mrtví 1. ukončující metoda
    def are_all_players_dead(self):
        for player in self.players:
            if player.is_alive:
                return False
        return True

    # Zkontroluje, zda vypršel časový limit generace 2. ukončující metoda
    @staticmethod
    def is_time_up():
        return current_time >= generation_time

    # Vrátí počet živých hráčů
    def get_alive_player_count(self):
        count = 0
        for player in self.players:
            if player.is_alive:
                count += 1
        return count

    def update_table(self, generation):
        # Přidání hodnot do tabulky pro aktuální generaci
        self.append = self.table.append(
            [generation, self.average_fitness, self.best_fitness, self.table_population_size])

    def display_table(self):
        # Výpis celé tabulky
        headers = ["Generace", "Průměrné fitness", "Nejlepší fitness", "Zbývající hráčí"]
        print(tabulate(self.table, headers=headers))


# Inicializace aplikace PyQt pro vykreslování statistik
app = QApplication(sys.argv)
# Vytvoření instance okna statistiky
stats_app = QApplication(sys.argv)
# Inicializace Pygame
pygame.init()
window = pygame.display.set_mode((window_width, window_height))
pygame.display.set_caption("Příklad evolučních algoritmů")
# Vytvoření cíle
target_position = (window_width // 2, 50)
target_size = 50
target = Target(target_position, target_size)
# Inicializace populace
population_size = 1000
elite_percentage = 0.1
gene_length = 5000  # Počet genů v chromozomu (délka kódu pohybu hráče)
# Inicializace okna pro nastavení simulace
dialog = ConfigurationDialog()
# Interakce s dialogovým oknem při spuštění aplikace - výběr křížení, velikost populace, selekce a druhu překážek
if dialog.exec() == QDialog.Accepted:
    crossover_type = dialog.get_crossover_type()
    population_size = dialog.get_population_size()
    selection_type = dialog.get_selection_type()
    obstacle_type = dialog.get_obstacle_type()
else:
    sys.exit(0)
# Nastavení překážek
obstacles = []

if obstacle_type == "Jednoduchá překážka":
    obstacles.append(Obstacle(350, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(400, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(450, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(250, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(200, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(500, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(550, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(300, 300, 50, 50, obstacle_type))
elif obstacle_type == "Složité překážky":
    obstacles.append(Obstacle(350, 150, 50, 50, obstacle_type))
    obstacles.append(Obstacle(450, 150, 50, 50, obstacle_type))
    obstacles.append(Obstacle(300, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(250, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(200, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(500, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(550, 300, 50, 50, obstacle_type))
    obstacles.append(Obstacle(100, 200, 50, 50, obstacle_type))
    obstacles.append(Obstacle(50, 200, 50, 50, obstacle_type))
    obstacles.append(Obstacle(600, 200, 50, 50, obstacle_type))
    obstacles.append(Obstacle(650, 200, 50, 50, obstacle_type))
else:
    raise ValueError("Neplatný druh překážek")
# Vytvoření hranice
border_obstacles = Obstacle.create_border_obstacles()
# Výpočet minimální a maximální vzdálenosti od cíle
min_distance = 0  # Nejkratší vzdálenost je 0, pokud hráč dosáhl cíle
max_distance = math.sqrt(window_width ** 2 + window_height ** 2)  # Nejdelší vzdálenost je úhlopříčka okna
population = Population(population_size, elite_percentage, gene_length, None, min_distance, max_distance)
stats_window = StatsWindow(population)
population.start_game(crossover_type, population_size, selection_type)
# Vytvoření tabulky
population.display_table()


# Vytvoření a uložení teplotní mapy do formátu PNG
def create_heatmap_and_save_to_png(players, window_width, window_height, output_file, target_position=None,
                                   obstacles=None):
    if obstacles is None:
        obstacles = []
    heatmap_grid = np.zeros((window_height + 25, window_width + 25),
                            dtype=int)  # Vytvoření prázdné mřížky s rozměry rozšířenými o 25 pixelů
    for player in players:
        for player_pos in player.positions_history:  # Projití pozic hráčů uložené v seznamu
            x_index = round(player_pos[0]) + 25  # Posunutí indexů
            y_index = round(player_pos[1]) + 25
            if 0 <= x_index < window_width + 25 and 0 <= y_index < window_height + 25:
                heatmap_grid[y_index][x_index] += 70
    if target_position:  # Zobrazení cíle na teplotní mapě
        target_x = round(target_position[0]) + 25
        target_y = round(target_position[1]) + 25
        if 0 <= target_x < window_width + 25 and 0 <= target_y < window_height + 25:
            target_region_size = 20
            for y in range(target_y - target_region_size // 2, target_y + target_region_size // 2 + 1):
                for x in range(target_x - target_region_size // 2, target_x + target_region_size // 2 + 1):
                    if 0 <= x < window_width + 25 and 0 <= y < window_height + 25:
                        heatmap_grid[y][x] = 100
            heatmap_grid[target_y][target_x] = 50  # Znázornění cílové pozice
    for obstacle in obstacles:  # Zobrazení překážek na teplotní mapě
        obstacle_x1 = int(obstacle.x - obstacle.width / 2) + 25
        obstacle_y1 = int(obstacle.y - obstacle.height / 2) + 25
        obstacle_x2 = int(obstacle.x + obstacle.width / 2) + 25
        obstacle_y2 = int(obstacle.y + obstacle.height / 2) + 25
        for y in range(obstacle_y1, obstacle_y2 + 1):
            for x in range(obstacle_x1, obstacle_x2 + 1):
                if 0 <= x < window_width + 25 and 0 <= y < window_height + 25:
                    heatmap_grid[y][x] = 200
    heatmap_grid = np.clip(heatmap_grid, 0, 1000)  # omezení hodnot na 0 až 1000
    plt.figure(figsize=(8, 6))
    plt.imshow(heatmap_grid[25:-25, 25:-25], cmap='cividis', interpolation='None')  # Vykreslení bez okrajů
    plt.title('Teplotní mapa pozic hráčů', fontsize=16)
    plt.colorbar(label='Počet hráčů', shrink=0.8)
    plt.axis('off')
    plt.savefig(output_file, bbox_inches='tight', pad_inches=0.1, dpi=300)
    plt.close()


# Herní smyčka
running = True
target_fps = 60
generation_time = 30
# Časový limit pro každou generaci (v sekundách)
clock = pygame.time.Clock()
current_time = 0  # Aktuální čas
while running:
    # Aktualizace času
    dt = clock.tick(target_fps)  # Získání uplynulého času v milisekundách
    dt_seconds = dt / 1000  # Převod na sekundy

    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            running = False
        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_SPACE:
                # Zrychlení času a rychlostí hráčů při stisknutí mezerníku
                for player in population.players:
                    target_fps = 240  # Zdvojnásobení rychlosti hráčů
            elif event.key == pygame.K_ESCAPE:  # Ukončení hry při stisknutí klávesy Escape
                running = False
                sys.exit()
        elif event.type == pygame.KEYUP:
            if event.key == pygame.K_SPACE:
                # Návrat rychlostí hráčů na původní hodnoty po puštění mezerníku
                target_fps = 60
    # Vykreslení okna pro statistiky
    stats_window.update()
    # Spuštění PyQt aplikace pro zobrazení okna statistik
    stats_app.processEvents()
    # Vykreslení pozadí
    background_image_scaled = pygame.transform.scale(background_image, (window_width, window_height))
    window.blit(background_image_scaled, (0, 0))
    # Update a vykreslování populace
    population.update()
    population.draw()
    # Vykreslení cíle
    target.draw()
    # Vykreslení překážek
    for obstacle in obstacles:
        obstacle.draw()
    # Vykreslení hranice
    for border_obstacle in border_obstacles:
        border_obstacle.draw()
    pygame.display.update()
    # Pokud se hráči již pohnuli, začít počítat čas
    if target_fps == 60:
        current_time += dt_seconds
    else:
        current_time += dt_seconds * (target_fps / 60)
    # 1. a 2. kontrola ukončení generace
    if population.are_all_players_dead() or population.is_time_up():
        population.table_population_size = population.get_alive_player_count()
        if population.get_alive_player_count() == 0:  # Pokud nikdo nepřežil v předchozí generaci
            # Resetování hry
            population = Population(population_size, elite_percentage, gene_length, stats_window, min_distance,
                                    max_distance)
            target.reset()  # Resetování pozice cíle
            for obstacle in obstacles:
                obstacle.x = obstacle.initial_x  # Resetování pozice překážky
                obstacle.y = obstacle.initial_y
            for border_obstacle in border_obstacles:
                border_obstacle.x = border_obstacle.initial_x  # Resetování pozice hranice
                border_obstacle.y = border_obstacle.initial_y
            current_time = 0  # Resetování času
        else:
            population.next_generation()
            current_time = 0  # Resetování času
    # Pokud se hráč dostane do cíle, tak se volá metoda game_over, která končuje hru
    if any(player.is_finished for player in population.players):
        population.players[0].game_over()
        running = False
pygame.quit()
app.quit()
